/*=========================================================================
 Authors: The GoFigure Dev. Team.
 at Megason Lab, Systems biology, Harvard Medical school, 2009
 
 Copyright (c) 2009, President and Fellows of Harvard College.
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
 
 Redistributions of source code must retain the above copyright notice,
 this list of conditions and the following disclaimer.
 Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation
 and/or other materials provided with the distribution.
 Neither the name of the  President and Fellows of Harvard College
 nor the names of its contributors may be used to endorse or promote
 products derived from this software without specific prior written
 permission.
 
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS
 BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
 OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
 ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
 =========================================================================*/

#ifndef __itkFastICAImageFilter_h
#define __itkFastICAImageFilter_h

#include <itkImageToImageFilter.h>

#include <vnl/vnl_matrix.h>
#include <vnl/vnl_vector.h>
#include <vnl/algo/vnl_svd.h>
#include <vnl/vnl_random.h>

namespace itk
{
/**
 * \class FastICAImageFilter
 * \brief This filter estimates the independent components from given images.
 * This code is based on Hyvarinen's fixed-point algorithm.
 * \author Arnaud Gelas arnaud_gelas@hms.harvard.edu
 * \author Kishore Mosaliganti kishore_mosaliganti@hms.harvard.edu
 * \date 10/21/2008
 */
template< class TInputImage, class TOutputImage >
class FastICAImageFilter :
  public ImageToImageFilter< TInputImage, TOutputImage >
{
public:
  typedef FastICAImageFilter Self;
  typedef SmartPointer< Self > Pointer;
  typedef SmartPointer< const Self > ConstPointer;
  typedef ImageToImageFilter< TInputImage, TOutputImage > Superclass;

  /** Run-time type information (and related methods).   */
  itkTypeMacro( FastICAImageFilter, ImageToImageFilter );
  /** New macro for creation of through a Smart Pointer   */
  itkNewMacro( Self );

  itkStaticConstMacro( ImageDimension, unsigned int,
    TInputImage::ImageDimension );

  typedef TInputImage InputImageType;
  typedef typename InputImageType::Pointer InputImagePointer;
  typedef typename InputImageType::PixelType InputImagePixelType;
  typedef typename InputImageType::SizeType InputImageSizeType;
  typedef typename InputImageType::SpacingType InputImageSpacingType;
  typedef typename InputImageType::PixelContainer InputImagePixelContainer;

  typedef TOutputImage OutputImageType;
  typedef typename OutputImageType::Pointer OutputImagePointer;
  typedef typename OutputImageType::SizeType OutputImageSizeType;
  typedef typename OutputImageType::SpacingType OutputImageSpacingType;
  typedef typename OutputImageType::PixelType OutputImagePixelType;
  typedef typename OutputImageType::PixelContainer OutputImagePixelContainer;
  typedef typename OutputImageType::PixelContainerPointer
    OutputImagePixelContainerPointer;

  typedef vnl_matrix< OutputImagePixelType > MatrixType;
  typedef vnl_vector< OutputImagePixelType > VectorType;

  enum InitializationType {
    Identity=0,
    Random
  };

  enum NonLinearFunctionType {
    CUBIC = 0,
    TANH
  };

  itkSetMacro( Initialization, InitializationType );
  itkSetMacro( NonLinearFunctionType, NonLinearFunctionType );

  itkSetMacro( SamplingRatio, double );
  itkGetMacro( SamplingRatio, double );

  itkSetMacro( NumberOfIterations, unsigned long );
  itkGetMacro( NumberOfIterations, unsigned long );

  itkSetMacro( Tolerance, OutputImagePixelType );
  itkGetMacro( Tolerance, OutputImagePixelType );

  itkSetMacro( W, MatrixType );
  itkGetMacro( W, MatrixType );

  itkGetMacro( A, MatrixType );

  itkSetMacro( Symmetric, bool );
  itkGetMacro( Symmetric, bool );

  itkSetMacro( Tanh_parameter, OutputImagePixelType );
  itkGetMacro( Tanh_parameter, OutputImagePixelType );

protected:
  FastICAImageFilter() : m_Initialization( Random ),
    m_NonLinearFunctionType( CUBIC ),
//     m_Stabilization( false ),
    m_Preprocessing( true ),
    m_Symmetric( false ),
    m_NumberOfInputs( 0 ),
    m_NumberOfPixels( 1 ),
    m_NumberOfSamples( 0 ), m_NumberOfIterations( 10000 ),
    m_SamplingRatio( 1. ), m_Random( 12 ), m_Tolerance( 1e-7 ),
    m_Tanh_parameter( 1. ), m_Threshold( 30 )
    {
    m_W.set_size(1,1);
    m_W[0][0]=0.;

    m_Temporary = m_W;
    }
  ~FastICAImageFilter()
  {}

  InitializationType m_Initialization;
  NonLinearFunctionType m_NonLinearFunctionType;
//   bool m_Stabilization;
  bool m_Preprocessing;
  bool m_Symmetric;
  unsigned int m_NumberOfInputs;
  unsigned long m_NumberOfPixels;
  unsigned int m_NumberOfSamples;
  unsigned long m_NumberOfIterations;

  double m_SamplingRatio;
  OutputImagePixelType m_MinAbsCos;

  vnl_random m_Random;
  VectorType m_Mean;

  /** \brief m_Data is a matrix of [NumberOfInputs x NumberOfPixels] */
  MatrixType m_Data;

  /** \brief m_X is a subsample of m_Data, a matrix of size
   * [NumberOfInputs x NumberOfSamples]
   */
  MatrixType m_X; // m_X = m_A * m_S
  MatrixType m_S; // m_S = m_W * m_X;
  MatrixType m_Whitening;
  MatrixType m_Dewhitening;
  MatrixType m_Temporary; // intermediate result
  MatrixType m_TemporaryOld; // intermediate result
  MatrixType m_A;
  MatrixType m_W;

  OutputImagePixelType m_Threshold;
  OutputImagePixelType m_Tolerance;
  OutputImagePixelType m_Tanh_parameter;

  typename DataObject::Pointer MakeOutput( unsigned int idx )
    {
//     if( idx < m_NumberOfComponents )
    if( idx < m_NumberOfInputs )
      {
      typename DataObject::Pointer output =
        ( OutputImageType::New() ).GetPointer();
      return output.GetPointer();
      }
    else
      {
      return 0;
      }
    }

  void Initialization()
    {
    m_NumberOfSamples =
      static_cast< unsigned long >( m_SamplingRatio * m_NumberOfPixels );

    m_Data.set_size( m_NumberOfInputs, m_NumberOfPixels );
    m_X.set_size( m_NumberOfInputs, m_NumberOfSamples );

    m_Whitening.set_size( m_NumberOfInputs, m_NumberOfInputs );
    m_Whitening.set_identity();
    m_Dewhitening = m_Whitening;

    m_Mean.set_size( m_NumberOfInputs );

    if( ( m_W.rows() < 2 ) && ( m_W.columns() < 2 ) )
      {
      if( m_Initialization == Identity )
        GenerateIdentityW();
      else
        {
        if( m_Initialization == Random )
          GenerateRandomW();
        }
      }
    else
      {
      m_Temporary = m_W;
      }

    m_TemporaryOld.set_size( m_Temporary.rows(), m_Temporary.columns() );
    m_TemporaryOld.fill( 0. );

    m_A = m_TemporaryOld;

    ConvertImageToVectors();
    }

  void GenerateData()
    {
    m_NumberOfInputs = this->GetNumberOfInputs();

    if( !IsValid() )
      return;

	std::cout << "Initialization." << std::endl;
    Initialization();

    if( m_Preprocessing )
	  {
	  std::cout << "PreProcessing." << std::endl;
      Preprocess( );
	  }

    if( m_Symmetric )
	  {
      std::cout << "SymetricApproach." << std::endl;
      SymmetricApproach();
	  }
    else
	  {
      std::cout << "DeflationApproach." << std::endl;
	  DeflationApproach();
	  }

    if( m_Preprocessing )
      PostProcess( );
    else
      m_S = m_W * m_Data;

    ConvertVectorsToImage( );
  }

  void SymmetricApproach( )
    {
    unsigned int k(0);
    do
      {
      RandomSubSampleOfData( );
      SymmetricDecorrelation( m_Temporary );
      ComputeError( );

      if( 1- m_MinAbsCos < m_Tolerance )
        {
        break;
        }

      m_TemporaryOld = m_Temporary;
      m_Temporary = EstimateNewMatrix( m_Temporary, m_X );

      std::cout <<"*** " <<k <<"***" <<std::endl;
      std::cout <<1 - m_MinAbsCos <<std::endl;
      std::cout <<"m_Temporary = " <<m_Temporary;
      ++k;
      } while( k < m_NumberOfIterations );

    m_W = m_Temporary.transpose() * m_Whitening;
    m_A = m_Dewhitening * m_Temporary;
    }


  void DeflationApproach()
    {
    OutputImagePixelType tol2 = m_Tolerance * m_Tolerance;

    m_W = m_Temporary;
    m_Temporary.fill( 0. );

    VectorType w_i;
    VectorType wOld( m_NumberOfInputs, 0. );

    for( unsigned int i = 0; i < m_NumberOfInputs; ++i )
      {
      // w is a vector of m_W
      w_i = m_W.get_column( i );
      w_i = w_i - m_Temporary * m_Temporary.transpose() * w_i;
      w_i.normalize();

      wOld.fill( 0. );

      for( unsigned int j = 0; j < m_NumberOfIterations; ++j )
        {
        RandomSubSampleOfData( );

        w_i = w_i - m_Temporary * m_Temporary.transpose() * w_i;
        w_i.normalize();

        std::cout <<"** " <<i <<" * " <<j <<" [" <<w_i <<"] " <<( w_i - wOld
).two_norm() <<std::endl;

        if( ( ( w_i - wOld ).two_norm() < tol2 ) ||
            ( ( w_i + wOld ).two_norm() < tol2 ) )
          {
          m_Temporary.set_column( i, w_i );

          VectorType u = m_Dewhitening * w_i;
          m_A.set_column( i, u );

          u = m_Whitening.transpose() * w_i;
          m_W.set_row( i, u );
          break;
          }
        if( j == m_NumberOfIterations - 1 )
          {
          m_Temporary.set_column( i, w_i );

          VectorType u = m_Dewhitening * w_i;
          m_A.set_column( i, u );

          u = m_Whitening.transpose() * w_i;
          m_W.set_row( i, u );
          break;
          }

        wOld = w_i;

        w_i = EstimateNewVector( w_i );
        }
      }
    }

  VectorType EstimateNewVector( const VectorType& iW )
    {
    VectorType oW;
    unsigned long kk;
    OutputImagePixelType
      inv_N = 1. / static_cast< OutputImagePixelType >( m_NumberOfSamples );

    VectorType Y = m_X.transpose() * iW;

    switch( m_NonLinearFunctionType )
      {
      default:
      case CUBIC:
        for( kk = 0; kk < Y.size(); ++kk )
          Y[kk] = Y[kk] * Y[kk] * Y[kk] * inv_N;

        oW = ( m_X * Y ) - 3. * iW;
        break;
      case TANH:
        OutputImagePixelType sum = 0.;

        for( kk = 0; kk < Y.size(); ++kk )
          {
          Y[kk] = tanh( m_Tanh_parameter * Y[kk] ) * inv_N;
          sum += 1 - Y[kk] * Y[kk];
          }
          sum *= inv_N;
        oW = ( m_X * Y - m_Tanh_parameter * sum * iW );
        break;
      }

    oW.normalize();
    return oW;
    }
  /**
   * \brief Check that input images have same size, same spacing
   */
  bool IsValid( )
  {
    InputImageSizeType size0 =
      this->GetInput( 0 )->GetLargestPossibleRegion().GetSize();
    InputImageSpacingType spacing0 = this->GetInput( 0 )->GetSpacing();
    m_NumberOfPixels = 1.;
    unsigned int dim( 0 ), i( 1 );

    for( ; dim < ImageDimension; ++dim )
      m_NumberOfPixels *= size0[dim];

    for( ; i < m_NumberOfInputs; ++i )
      {
      InputImageSizeType size =
        this->GetInput( i )->GetLargestPossibleRegion().GetSize();

      InputImageSpacingType spacing =
        this->GetInput( i )->GetSpacing( );

      for( dim = 0; dim < ImageDimension; ++dim )
        {
        if( size[dim] != size0[dim] )
          {
          std::cout <<"Images of different size" <<std::endl;
          return false;
          }
        if( spacing[dim] != spacing0[dim] )
          {
          std::cout <<"Images of different spacing" <<std::endl;
          return false;
          }
        }
      }
    return true;
  }

  /**
   * \brief Generate m_X from m_Data, i.e. a random subsample of input images
   */
  void RandomSubSampleOfData()
    {
    unsigned long id;
    unsigned int j;

    std::map< unsigned long, bool > map_id;

    if( m_NumberOfSamples != m_NumberOfPixels )
      {
      for( unsigned long i = 0; i < m_NumberOfSamples; ++i )
        {
        do
          {
          id = m_Random.lrand32( 0, m_NumberOfPixels );
          } while( map_id.find( id ) != map_id.end() );

        map_id[id]=true;

        for( j = 0; j < m_NumberOfInputs; ++j )
          m_X[j][i] = m_Data[j][id];
        }
      }
    else
      m_X = m_Data;
    }

  void GenerateIdentityW()
    {
    m_Temporary.set_size( m_NumberOfInputs, m_NumberOfInputs );
    m_Temporary.set_identity();
    }

  void GenerateRandomW( )
    {
    m_Temporary.set_size( m_NumberOfInputs, m_NumberOfInputs );

    unsigned int i, j;

    for( i = 0; i < m_NumberOfInputs; ++i )
      {
      for( j = i; j < m_NumberOfInputs; ++j )
        {
        m_Temporary[i][j] = m_Random.drand32( -1., 1. );
        m_Temporary[j][i] = m_Random.drand32( -1., 1. );
        }
      }

    if( m_Symmetric )
      SymmetricDecorrelation( m_Temporary );
    else
      {
      VectorType w_i = m_Temporary.get_column( 0 );
      w_i.normalize();
      m_Temporary.set_column( 0, w_i );

      VectorType w_j;
      for( i = 1; i < m_NumberOfInputs; ++i )
        {
        w_i = m_Temporary.get_column( i );
        for( j = 0; j < i; ++j )
          {
          w_j = m_Temporary.get_column( j );
          w_i = w_i - dot_product( w_i, w_j );
          }
        w_i.normalize();
        m_Temporary.set_column( i, w_i );
        }
      }
    }

  void ConvertImageToVectors()
    {
    unsigned int i;
    unsigned long kk;

    for( i = 0; i < m_NumberOfInputs; ++i )
      {
      InputImageType* input =
        const_cast< InputImageType* >( this->GetInput( i ) );
      InputImagePixelContainer* container = input->GetPixelContainer();
      InputImagePixelType* data = container->GetImportPointer();

      for( kk = 0; kk < m_NumberOfPixels; ++kk )
        {
        m_Data[i][kk] = static_cast< OutputImagePixelType >( data[kk] );
        }
      }
    }

  void ConvertVectorsToImage( )
    {
    this->SetNumberOfRequiredOutputs( m_NumberOfInputs );

    unsigned int i;
    unsigned long kk;

    for( i = 0; i < m_NumberOfInputs; ++i )
      {
      this->SetNthOutput( i, MakeOutput( i ) );
      OutputImagePointer output = this->GetOutput( i );
      output->SetBufferedRegion( this->GetInput( i )->GetBufferedRegion() );
      output->Allocate();
      VectorType v = m_S.get_row(i);
      for( kk = 0; kk < v.size(); ++kk )
        {
        output->GetPixelContainer()->GetImportPointer()[kk] = v[kk];
        }
      }
    }

  /** \brief Center and Whiten m_Data */
  void Preprocess( )
    {
    VectorType v;
    // first centering
    for( unsigned int i = 0; i < m_NumberOfInputs; ++i )
      {
      v = m_Data.get_row( i );
      m_Mean[i] = v.mean();
      m_Data.set_row( i, v - m_Mean[i] );
      }

    // Then whiten and computes Dewhitening matrix
    Whitening( m_Data );
    m_Temporary = m_Whitening * m_Temporary;
    }

  /** \brief Unwhiten and uncenter */
  void PostProcess( )
    {
    m_S = m_W * m_Data;
    VectorType new_mean = m_W * m_Mean;

    for( unsigned int i = 0; i < m_NumberOfInputs; ++i )
      {
      m_S.set_row( i, m_S.get_row( i ) + new_mean[i] );
      }
    }

  /**
   * \brief Compute the Covariance Matrix C=ioX'*ioX/size(ioX).
   * SVD decomposition of C, i.e.
   * \f[C=svd.U()\cdot svd.W()\cdot svd.V()^t\f]
   * \f[ Whitening = ( \sqrt( svd.W() )*svd.V()^t ) )^{-1}\f]
   * \f[ Dewhitening = svd.U() * sqrt( svd.W() )\f]
   * \f[ioX = Whitening * ioX\f]
   * \param[in,out] ioX
   */
  void Whitening( MatrixType& ioX )
    {
    MatrixType covariance = ioX * ioX.transpose()
      / static_cast< OutputImagePixelType >( ioX.columns() );

    vnl_svd< OutputImagePixelType > svd( covariance );
    svd.zero_out_absolute( 1e-16 );

    MatrixType sigma( covariance.rows(), covariance.columns(), 0. );
    MatrixType inv_sigma( covariance.rows(), covariance.columns(), 0. );

    for( unsigned int i = 0; i < covariance.rows(); ++i )
      {
      sigma[i][i] = vcl_sqrt( svd.W( i ) );
      inv_sigma[i][i] = 1. / sigma[i][i];
      }
    m_Whitening = inv_sigma * svd.V().transpose();
    m_Dewhitening = svd.U() * sigma;
    ioX = m_Whitening * ioX;
    }

  /**
   * \param[in] iX Matrix of size [NumberOfInputs x NumberOfSamples]
   * \param[in] iW Matrix of size [NumberOfInputs x NumberOfInputs]
   */
  MatrixType EstimateNewMatrix( const MatrixType& iW,
    const MatrixType& iX )
  {
    // size(Y) = [NumberOfSamples x NumberOfInputs]
    MatrixType Y = iX.transpose() * iW;
    unsigned long i, j;
    OutputImagePixelType inv_N =
      1. / static_cast< OutputImagePixelType >( m_NumberOfSamples );

    switch( m_NonLinearFunctionType )
      {
      default:
      case CUBIC:
        for( i = 0; i < Y.columns(); ++i )
          {
          for( j = 0; j < Y.rows(); ++j )
            {
            Y[i][j] = Y[i][j] * Y[i][j] * Y[i][j] * inv_N;
            }
          }
        return iX * Y  - 3. * iW;

      case TANH:
        MatrixType Mat( iW.rows(), iW.columns() );

        OutputImagePixelType temp, sum;

        for( i = 0; i < Y.columns(); ++i )
          {
          sum = 0.;
          for( j = 0; j < Y.rows(); ++j )
            {
            temp = vcl_tanh( m_Tanh_parameter * Y[i][j] );
            Y[i][j] = temp * inv_N;
            sum += 1 - temp * temp;
            }
          sum *= m_Tanh_parameter;

          for( j = 0; j < m_Temporary.columns(); ++j )
            Mat[i][j] = sum * m_Temporary[i][j] * inv_N;
          }

        return iX * Y - Mat;
      }

  }

  void ComputeError( )
  {
    MatrixType u = m_Temporary * m_TemporaryOld;

    m_MinAbsCos = NumericTraits< OutputImagePixelType >::max();

    for( unsigned int i = 0; i < m_Temporary.rows(); ++i )
    {
      m_MinAbsCos = vnl_math_min( vnl_math_abs( u[i][i] ),
        m_MinAbsCos );
    }
  }

  void SymmetricDecorrelation( MatrixType& ioW )
  {
    MatrixType B = ioW.transpose() * ioW;
    vnl_svd< OutputImagePixelType > svd( B );
    svd.zero_out_absolute( 1e-16 );

    MatrixType Sigma( B.rows(), B.rows(), 0. );

    for( unsigned int i = 0; i < B.rows(); ++i )
      {
      Sigma[i][i] = 1. / vcl_sqrt( svd.W( i ) );
      }

    ioW = ioW * svd.V() * Sigma * svd.U().transpose();
  }

private:
  FastICAImageFilter( const Self& );
  void operator = ( const Self& );
};
}
#endif
